"""
A FastAPI HTML server.

The streamlit version had issues around state management and was genrally slow
and inflexible. This gives us a lot more control.
"""
import datetime
import importlib
import json
import logging
import random
import string
from typing import Annotated, Type

import llm
import requests
from fastapi import FastAPI, Form, HTTPException, Request, responses, staticfiles, templating

from fossil_mastodon import algorithm, config, core, migrations, plugins, ui


logger = logging.getLogger(__name__)


app = FastAPI(lifespan=plugins.lifespan)


app.mount("/static", staticfiles.StaticFiles(directory=config.ASSETS.assets_path), name="static")
templates = templating.Jinja2Templates(directory=config.ASSETS.templates_path)
print("using template directory", config.ASSETS.templates_path)
templates.env.filters["rel_date"] = ui.time_ago


@app.middleware("http")
async def session_middleware(request: Request, call_next):
    """
    Called before each request. Sets up the session and saves it to the database.
    """
    session_id = request.cookies.get("fossil_session_id")
    session = core.Session.get_by_id(session_id) if session_id else None
    if session is None:
        session = core.Session.get_or_create()
        session.save()
        request.state.session = session
        response = await call_next(request)
        response.set_cookie("fossil_session_id", session.id)  
        return response
    else:
        request.state.session = session
        return await call_next(request)


@app.get("/")
async def root(request: Request):
    session: core.Session = request.state.session
    ctx = plugins.RenderContext(
        templates=templates,
        request=request,
        link_style=ui.LinkStyle("Desktop"),
        session=session,
    )

    # GUARD: ensure some algorithms are installed
    algo_list = plugins.get_algorithms()
    if len(algo_list) == 0:
        print(f"No algorithms found (num plugins: {len(plugins.get_plugins())})")
        for plugin in plugins.get_plugins():
            print(f"Plugin ({plugin.name})", plugin)
        return templates.TemplateResponse("no_algorithm.html", {
            "request": request,
        })

    # Render the UI
    algo = session.get_algorithm_type() or algo_list[0]
    return templates.TemplateResponse("index.html", {
        "request": request,
        "model_params": algo.render_model_params(ctx).body.decode("utf-8"),
        "ui_settings": session.get_ui_settings(),
        "selected_algorithm": algo,
        "algorithms": [
            {"name": algo.plugin.name, "display_name": algo.plugin.display_name}
            for algo in plugins.get_algorithms()
        ],
    })


@app.get("/toots")
async def toots():
    return staticfiles.FileResponse("public/toots.html")


@app.post("/toots/download")
async def toots_download(request: Request):
    # init
    migrations.create_database()
    session: core.Session = request.state.session
    algorithm_spec: dict = json.loads(session.algorithm_spec) if session.algorithm_spec else {}

    # first page load calls this with display-only=true to load what was loaded last time
    if request.query_params.get("display-only", "") != "true":
        # download
        core.download_timeline(datetime.datetime.utcnow() - datetime.timedelta(days=1), session.id)

    # render
    body_params: dict[str, str] = dict((await request.form()))
    session.set_ui_settings(body_params)
    print("algorithm_spec", algorithm_spec)
    if "module" in algorithm_spec and "class_name" in algorithm_spec:
        mod = importlib.import_module(algorithm_spec["module"])
        model_class: Type[algorithm.BaseAlgorithm] = getattr(mod, algorithm_spec["class_name"])
        model: algorithm.BaseAlgorithm = model_class.deserialize(session.algorithm)
        timespan = ui.timedelta(body_params["time_span"])
        timeline = core.Toot.get_toots_since(datetime.datetime.utcnow() - timespan)
        renderable = model.render(timeline, plugins.RenderContext(
            templates=templates,
            request=request,
            link_style=ui.LinkStyle(body_params["link_style"] if "link_style" in body_params else "Desktop"),
            session=session,
        ))
        return renderable.render()
    else:
        return responses.HTMLResponse("<div>No Toots 😥</div>")


@app.post("/toots/train")
async def toots_train(
    link_style: Annotated[str, Form()],
    time_span: Annotated[str, Form()],
    request: Request,
):
    context = algorithm.TrainContext(
        end_time=datetime.datetime.utcnow(),
        timedelta=ui.timedelta(time_span),
        session_id=request.state.session.id
    )

    algo_kwargs = {k: v for k, v in dict((await request.form())).items() 
                   if k not in {"link_style", "time_span"}}
    print("Algorithm kwargs:", algo_kwargs)

    # train
    session: core.Session = request.state.session
    algo = session.get_algorithm_type() or plugins.get_algorithms()[0]
    algo.model_version = "".join(random.choices(string.ascii_letters + string.digits, k=12))
    model = algo.train(context, algo_kwargs)
    session.algorithm = model.serialize()
    session.algorithm_spec = json.dumps({
        "module": model.__class__.__module__,
        "class_name": model.__class__.__qualname__,
        "kwargs": algo_kwargs,
    })
    session.save()

    # render
    timeline = core.Toot.get_toots_since(datetime.datetime.utcnow() - ui.timedelta(time_span))
    renderable = model.render(timeline, plugins.RenderContext(
        templates=templates,
        request=request,
        link_style=ui.LinkStyle(link_style),
        session=session,
    ))
    try:
        return renderable.render()
    except plugins.BadPluginFunction as ex:
        return templates.TemplateResponse("bad_plugin.html", { "request": request, "ex": ex })


@app.get("/algorithm/{name}/form")
async def algorithm_form(name: str, request: Request):
    session: core.Session = request.state.session
    algo_type = session.get_algorithm_type() or plugins.get_algorithms()[0]
    ctx = plugins.RenderContext(
        templates=templates,
        request=request,
        link_style=ui.LinkStyle(session.get_ui_settings().get("link_style", "Desktop")),
        session=session,
    )
    return algo_type.render_model_params(ctx)


@app.get("/settings")
async def get_settings(request: Request):
    session: core.Session = request.state.session
    keys = {"openai": "", **llm.load_keys()}
    return templates.TemplateResponse("settings.html", {
        "request": request,
        "settings": session.settings,
        "embedding_models": config.get_installed_embedding_models(),
        "embedding_model": session.settings.embedding_model,
        "summarize_models": config.get_installed_llms(),
        "summarize_model": session.settings.summarize_model,
        "keys": keys,
    })

@app.post("/settings")
async def post_settings(settings: core.Settings, request: Request):
    session: core.Session = request.state.session
    session.settings = settings
    session.save()
    return responses.HTMLResponse("<div>👍</div>")

@app.post("/keys")
async def post_keys(request: Request):
    body_params: dict[str, str] = dict((await request.form()))
    key_path = llm.user_dir() / "keys.json"
    key_path.write_text(json.dumps(body_params))
    return responses.HTMLResponse("<div>👍</div>")


@app.post("/toots/{id}/debug")
async def toots_debug(id: int):
    toot = core.Toot.get_by_id(id)
    if toot is not None:
        import json
        print(json.dumps(toot.orig_dict, indent=2))
    return responses.HTMLResponse("<div>💯</div>")

@app.post("/toots/{id}/boost")
async def toots_boost(id: int):
    toot = core.Toot.get_by_id(id)
    if toot is not None:
        url = f'{config.ConfigHandler.MASTO_BASE}/api/v1/statuses/{toot.toot_id}/reblog'
        data = {
            'visibility': 'public'
        }
        response = requests.post(url, json=data, headers=config.headers())
        try:
            response.raise_for_status()
            return responses.HTMLResponse("<div>🚀</div>")
        except:
            print("ERROR:", response.json())
            raise
    raise HTTPException(status_code=404, detail="Toot not found")

@app.post("/toots/{id}/favorite")
async def toots_favorite(id: int):
    toot = core.Toot.get_by_id(id)
    if toot is not None:
        url = f'{config.ConfigHandler.MASTO_BASE}/api/v1/statuses/{toot.toot_id}/favourite'
        response = requests.post(url, headers=config.headers())
        try:
            response.raise_for_status()
            return responses.HTMLResponse("<div>💫</div>")
        except:
            print("ERROR:", response.json())
            raise
    raise HTTPException(status_code=404, detail="Toot not found")
    
templates.env.globals["extra_menu_items"] = plugins.get_menu_items
templates.env.globals["head_html"] = plugins.get_head_html
templates.env.globals["extra_nav"] = plugins.get_extra_nav

# this should always be the last line of this file
plugins.init_plugins(app)


"""
Migration scripts to update the SQLite schema that are run at the last possible moment.
There's no version numbers, so each script is responsible for "knowing" when it needs
to run itself.

Typically you should have a @lru_cache on each function to prevent unnecessary invocations,
but also know that it'll get re-invoked every time the server restarts.
"""
import functools
import random
import sqlite3
import string

from fossil_mastodon import config

class migration:
    """
    Decorator that tracks all migration functions.
    """
    all: list["migration"] = []
    __counter = 0

    def __init__(self, func: callable):
        self.func = func
        self.cached = functools.lru_cache()(func)
        migration.__counter += 1
        self.id = migration.__counter
        migration.all.append(self)

    def __call__(self, *args, **kwargs):
        return self.cached(*args, **kwargs)


@migration
def create_database():
    with config.ConfigHandler.open_db() as conn:
        c = conn.cursor()

        # Create the toots table if it doesn't exist
        c.execute('''
            CREATE TABLE IF NOT EXISTS toots (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                content TEXT,
                author TEXT,
                url TEXT,
                created_at DATETIME,
                embedding BLOB,
                orig_json TEXT,
                cluster TEXT  -- Added cluster column
            )
        ''')

        conn.commit()


@migration
def create_session_table():
    create_database()
    with config.ConfigHandler.open_db() as conn:
        c = conn.cursor()

        # Create the toots table if it doesn't exist
        c.execute('''
            CREATE TABLE IF NOT EXISTS sessions (
                id TEXT PRIMARY KEY,
                algorithm_spec TEXT,
                algorithm BLOB,
                ui_settings TEXT
            )
        ''')

        try:
            c.execute('''
                ALTER TABLE sessions ADD COLUMN settings TEXT
            ''')
        except sqlite3.OperationalError:
            pass

        # Add session name
        try:
            c.execute('''
                ALTER TABLE sessions ADD COLUMN name TEXT
            ''')
        except sqlite3.OperationalError:
            pass

        c.execute("DELETE FROM sessions WHERE name IS NULL")

        c2 = conn.cursor()
        c2.execute("SELECT COUNT(*) FROM sessions")
        row_count = c2.fetchone()[0]
        if row_count == 0:
            rand_str = "".join(random.choice(string.ascii_lowercase) for _ in range(32))
            c2.execute("""
            INSERT INTO sessions (id, name, settings)
                VALUES (?, ?, '{}')
            """, (rand_str, "Main"))

        conn.commit()


import functools
import random
import string
import llm
import numpy as np
import pydantic
import tiktoken
from fastapi import Response, responses
from sklearn.cluster import KMeans
from tqdm import trange

from fossil_mastodon import algorithm, config, core, migrations, plugins, ui


plugin = plugins.Plugin(
    name="Topic Cluster",
    description="Cluster toots by topic",
)


class ClusterRenderer(algorithm.Renderable, pydantic.BaseModel):
    clusters: list[ui.TootCluster]
    context: plugins.RenderContext

    def render(self, **response_args) -> Response:
        toot_clusters = ui.TootClusters(clusters=self.clusters)
        return self.context.templates.TemplateResponse("toot_clusters.html", {
            "clusters": toot_clusters,
            **self.context.template_args(),
        },
        **response_args)


@migrations.migration
def _create_table():
    with config.ConfigHandler.open_db() as conn:
        c = conn.cursor()

        # Create the toots table if it doesn't exist
        c.execute('''
            CREATE TABLE IF NOT EXISTS topic_cluster_toots (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                toot_id INTEGER NOT NULL,
                model_version TEXT NOT NULL,
                cluster_id INTEGER NOT NULL,
                updated_at DATETIME DEFAULT CURRENT_TIMESTAMP
            )
        ''')

        conn.commit()


class TootModel(pydantic.BaseModel):
    """
    Cache for the cluster id of a toot. The model_version is used to invalidate the cache if 
    the model is retrained, since that would lead to an incompatible set of clusters.

    We can't store this inside the model because it's dynamic and created after the model is
    trained.
    """
    id: int | None
    toot_id: int
    model_version: str
    cluster_id: int | None

    @classmethod
    def for_toots(cls, toots: list[core.Toot], model_version: str) -> list["TootModel"]:
        _create_table()
        with config.ConfigHandler.open_db() as conn:
            c = conn.cursor()
            c.execute('''
                SELECT id, toot_id, model_version, cluster_id
                FROM topic_cluster_toots
                WHERE model_version = ?
            ''', (model_version, ))
            from_db = {row[1]: cls(id=row[0], toot_id=row[1], model_version=row[2], cluster_id=row[3]) for row in c.fetchall()}
            return [
                from_db.get(
                    toot.id, 
                    cls(id=None, toot_id=toot.id, model_version=model_version, cluster_id=None),
                ) 
                for toot in toots
            ]

    def save(self):
        _create_table()
        if self.cluster_id is None:
            raise ValueError("Cannot save a toot model without a cluster_id")

        if isinstance(self.cluster_id, np.number):
            raise ValueError("cluster_id must be an int, not a numpy type")

        with config.ConfigHandler.open_db() as conn:
            c = conn.cursor()
            if self.id is None:
                c.execute('''
                    INSERT INTO topic_cluster_toots (toot_id, model_version, cluster_id)
                    VALUES (?, ?, ?)
                ''', (self.toot_id, self.model_version, self.cluster_id))
                self.id = c.lastrowid
            else:
                c.execute('''
                    UPDATE topic_cluster_toots
                    SET cluster_id = ?, updated_at = CURRENT_TIMESTAMP
                    WHERE id = ?
                ''', (self.cluster_id, self.id))
            conn.commit()


@plugin.algorithm
class TopicCluster(algorithm.BaseAlgorithm):
    def __init__(self, kmeans: KMeans, labels: dict[int, str], model_version: str | None = None):
        self.kmeans = kmeans
        self.labels = labels
        self.model_version = model_version

    def render(self, toots: list[core.Toot], context: plugins.RenderContext) -> ClusterRenderer:
        before = len(toots)
        toots = [toot for toot in toots if toot.embedding is not None]
        toot_models = TootModel.for_toots(toots, model_version=self.model_version)
        print("Removed", before - len(toots), "toots with no embedding (probably image-only).", f"{len(toots)} toots remaining.")
        
        # Assign clusters to the uncached toots
        unassigned = [toot for toot, toot_model in zip(toots, toot_models) if toot_model.cluster_id is None]
        if len(unassigned) > 0:
            unassigned_models = [toot_model for toot_model in toot_models if toot_model.cluster_id is None]
            cluster_indices = self.kmeans.predict(np.array([toot.embedding for toot in unassigned]))
            print(f"Assigning clusters for {len(unassigned)} toots; model_version={self.model_version}")
            for toot, cluster_index, toot_model in zip(unassigned, cluster_indices, unassigned_models):
                toot.cluster = self.labels[cluster_index]
                toot_model.cluster_id = int(cluster_index)
                toot_model.save()

        toot_clusters = ui.TootClusters(
            clusters=[
                ui.TootCluster(
                    id=i_cluster,
                    name=cluster_label,
                    toots=[toot for toot, toot_model in zip(toots, toot_models) if toot_model.cluster_id == i_cluster],
                )
                for i_cluster, cluster_label in self.labels.items()
            ]
        )
        return ClusterRenderer(clusters=toot_clusters.clusters, context=context)

    @classmethod
    def train(cls, context: algorithm.TrainContext, args: dict[str, str]) -> "TopicCluster":
        toots = [toot for toot in context.get_toots() if toot.embedding is not None]

        n_clusters = int(args["num_clusters"])
        if len(toots) < n_clusters:
            return cls(kmeans=NoopKMeans(n_clusters=1), labels={0: "All toots"})

        embeddings = np.array([toot.embedding for toot in toots])
        kmeans = KMeans(n_clusters=n_clusters)
        cluster_labels = kmeans.fit_predict(embeddings)

        labels: dict[int, str] = {}
        model = llm.get_model(config.ConfigHandler.SUMMARIZE_MODEL(context.session_id).name)
        for i_clusters in trange(n_clusters):
            clustered_toots = [toot for toot, cluster_label in zip(toots, cluster_labels) if cluster_label == i_clusters]
            combined_text = "\n\n".join([toot.content for toot in clustered_toots])

            # Use the summarizing model to summarize the combined text
            prompt = f"Create a single label that describes all of these related tweets, make it succinct but descriptive. The label should describe all {len(clustered_toots)} of these\n\n{combined_text}"
            summary = model.prompt(reduce_size(context.session_id, prompt)).text().strip()
            labels[int(i_clusters)] = summary

        model_version = "".join(random.choice(string.ascii_lowercase) for _ in range(12))
        return cls(kmeans=kmeans, labels=labels, model_version=model_version)

    @staticmethod
    def render_model_params(context: plugins.RenderContext) -> Response:
        default = context.session.get_ui_settings().get("num_clusters", "15")
        return responses.HTMLResponse(f"""
            <div class="slider">
                <input type="range" name="num_clusters" id="num_clusters" min="0" max="20" value="{default}" onchange="document.getElementById('num_clusters_value').innerHTML = this.value">
                <span><span class="slider-value" id="num_clusters_value">{default}</span> clusters</span>
            </div>
        """)

def get_encoding(session_id: str):
    try:
        return tiktoken.encoding_for_model(config.ConfigHandler.SUMMARIZE_MODEL(session_id).name)
    except KeyError:
        encoding_name = tiktoken.list_encoding_names()[-1]
        return tiktoken.get_encoding(encoding_name)

def reduce_size(session_id: str, text: str, model_limit: int = -1, est_output_size: int = 500) -> str:
    if model_limit < 0:
        model_limit = config.ConfigHandler.SUMMARIZE_MODEL(session_id).context_length
    tokens = get_encoding(session_id).encode(text)
    return get_encoding(session_id).decode(tokens[:model_limit - est_output_size])


class NoopKMeans(KMeans):
    def predict(self, X, y=None, sample_weight=None):
        return np.zeros(len(X), dtype=int)
